import sys

from ardi.driver import drivercore,driverbase
from pymodbus.client.sync import ModbusTcpClient as ModbusClient
try:    
    from pymodbus.payload import PayloadDecoder as BinaryPayloadDecoder
except:
    from pymodbus.payload import BinaryPayloadDecoder

from pymodbus.pdu import ExceptionResponse as ExceptionResponse

import logging

class fetchsetpoint:
    def __init__(self):
        self.point = 0
        self.offset = 0
        self.format = 0

class fetchset:
    def __init__(self):
        self.address = 0
        self.length = 0
        self.register = 0
        self.points = []
        
    def Setup(self):
        self.register = 0
        add = int(self.address)
        if add >= 100000:
            self.register = 1
        if add >= 300000:
            self.register = 3
        if add >= 400000:
            self.register = 4

        #print "**Register Type from " + str(self.address) + " = " + str(self.register)

    def Normalise(self, addr):
        if self.register == 4:
            return addr - 400000
        if self.register == 3:
            return addr - 300000
        if self.register == 1:
            return addr - 100000
        return addr

    def Extend(self,address,mode, code):
        len = 1
        if (mode >= 10):
            len = 2

        if self.address == -1:

            self.address = address
            self.length = len

            pnt = fetchsetpoint()
            pnt.point = code
            pnt.offset = 0
            pnt.format = mode
        
            self.points.append(pnt)
            self.Setup()
            return

        #print "Adding " + str(address)

        #print "   Start: " + str(self.address) + ", " + str(self.length) + " long";

        if (self.address > address):
            #print "   Extending Inwards"
            dist = self.address - address
            self.length = self.length + dist + len + 1
            self.address = address
            for pnt in self.points:
                pnt.offset += dist
        else:
            #print "   Extending Outwards"
            if self.address + self.length <= address + len:
                self.length = (address - self.address) + (len)

        #print "   End: " + str(self.address) + ", " + str(self.length) + " long";

        pnt = fetchsetpoint()
        pnt.point = code
        pnt.offset = address - self.address
        pnt.format = mode
        
        self.points.append(pnt)

class modbustcpdriver:
    def __init__(self):
        self.connected = False
        self.polling = True
        self.digits = 6
        self.endian = '>'
        self.fetches = []
    
    def SetAddress(self, addr):
        self.connected = False        
        bits = addr.split(":")
        self.ip = bits[0]
        if len(bits) > 1:
            self.port = int(bits[1])
        else:
            self.port = 502
        if len(bits) > 2:
            self.unit = int(bits[2])
        else:
            self.unit = 1
        if len(bits) > 3:
            self.endian = bits[3]
            if self.endian == '':
                self.endian = '>'
        if len(bits) > 4:
            self.digits = int(bits[4])
    
    def Connect(self):
        #self.SetAddress(addr)
        self.connected = False
        self.client = ModbusClient(self.ip, port=self.port)

        logging.basicConfig(level=logging.DEBUG)
        #print "Connecting To " + self.ip + "!"
        try:
            success = self.client.connect()
            if success == False:
                return False
            self.connected = True
            return True
        except:
           return False
    
    def Disconnect(self):
        self.connected = False
        self.client.close()
        
    def Poll(self):
        #print "Polling "+ str(len(self.core.pointlist)) + " items"
        #for item in self.core.pointlist:
        #    print "  > "+ item.name

        for fetch in self.fetches:
            requestbase = fetch.Normalise(fetch.address)
            if fetch.register == 0:
                rr = self.client.read_coils(requestbase,fetch.length,unit=self.unit)
                
                if type(rr) is ExceptionResponse:                    
                    continue

                q=-1
                for pnt in fetch.points:
                    q = q + 1
                    if rr.bits[q] == True:
                        v=1
                    self.core.UpdateData(pnt.code,v)
                    
                continue

            if fetch.register == 1:
                rr = self.client.read_discrete_inputs(requestbase,fetch.length,unit=self.unit)
                
                if type(rr) is ExceptionResponse:                    
                    continue

                q=-1
                for pnt in fetch.points:
                    q = q + 1
                    if rr.bits[q] == True:
                        v=1
                    self.core.UpdateData(pnt.code,v)
                    
                continue

            if fetch.register == 3:
                #print "Fetching " + str(fetch.length) + " from " + str(requestbase)
                rr = self.client.read_input_registers(requestbase,fetch.length,unit=self.unit)
            else:
                #print "**HOLDING REGISTERS?"
                rr = self.client.read_holding_registers(requestbase,fetch.length,unit=self.unit)            
                
            if type(rr) is ExceptionResponse:
                self.core.logger.warning("Error Detected: " + str(rr))
                continue

            if rr is None:
                self.core.logger.warning("Null Response Detected On Request For " + str(address-base))
                Disconnect()
                return

            for pnt in fetch.points:
                
                if pnt.format == 0:                
                    self.core.UpdateData(pnt.code,rr.registers[pnt.offset])
                    continue
            
                en = '>';
                decoder = BinaryPayloadDecoder.fromRegisters(rr.registers[pnt.offset:],endian=en)
                
                if pnt.format == 1:
                    val = decoder.decode_32bit_int()				
                    self.core.UpdateData(pnt.point,val)
                if pnt.format == 10:
                    val = decoder.decode_32bit_float()
                    self.core.UpdateData(pnt.point,val)
                    #print "Value = " + str(val)
                if pnt.format == 11:
                    val = decoder.decode_64bit_float()
                    self.core.UpdateData(pnt.point,val)
                if pnt.format == 20:
                    val = decoder.decode_string(int(bits[2]))
                    self.core.UpdateData(pnt.point,val)   
        return
    
        for item in self.core.pointlist:            
            bits = item.address.split('|')
            mode = 0
            address = 0
            
            if len(bits) == 1:
                address = int(bits[0])
            else:
                address = int(bits[0])
                mode = int(bits[1])

            base = 0
            range = 10000
            if self.digits == 6:
                range = 100000

            if address < range:
                rr = self.client.read_coils(address,1,unit=self.unit)
                v = 0
                if type(rr) is ExceptionResponse:                    
                    continue
                if rr.bits[0] == True:
                    v=1
                self.core.UpdateData(item.code,v)
                continue

            base = 10000
            range = 20000
            if self.digits == 6:
                base = 100000
                range = 200000

            if address < range:
                rr = self.client.read_discrete_inputs(address-base,1,unit=self.unit)
                v = 0                
                if type(rr) is ExceptionResponse:                    
                    continue
                if rr.bits[0] == True:
                    v=1
                self.core.UpdateData(item.code,v)
                continue

            ln = 1
            if (mode == 1):
                ln = 2
            elif (mode == 2):
                ln = 4
            elif (mode == 3):
                ln = int(bits[2])
            elif (mode == 10):
                ln = 4
            
            range = 40000
            if self.digits == 6:                
                range = 400000
            
            if address < range:
                base = 30000
                if self.digits == 6:
                    base = 300000
                                    
                rr = self.client.read_input_registers(address-base,ln,unit=self.unit)
            else:
                base = 40000
                if self.digits == 6:
                    base = 400000
                
                rr = self.client.read_holding_registers(address-base,ln,unit=self.unit)

            if type(rr) is ExceptionResponse:
                self.core.logger.warning("Error Detected: " + str(rr))
                continue

            if rr is None:
                self.core.logger.info("Null Response Detected On Request For " + str(address-base))
                Disconnect()
                return

            if mode == 0:                
                self.core.UpdateData(item.code,rr.registers[0])
                continue

            #logging.info("Processing Response...")

            en = '>';
            decoder = BinaryPayloadDecoder.fromRegisters(rr.registers,endian=en)
            
            if mode == 1:
                val = decoder.decode_32bit_int()				
                self.core.UpdateData(item.code,val)
            if mode == 10:
                val = decoder.decode_32bit_float()
                self.core.UpdateData(item.code,val)                
            if mode == 11:
                val = decoder.decode_64bit_float()
                self.core.UpdateData(item.code,val)
            if mode == 20:
                val = decoder.decode_string(int(bits[2]))
                self.core.UpdateData(item.code,val)            

    def Optimise(self):

        #print "-------"
        self.fetches = []
        overhead = 16
        for item in self.core.pointlist:
            found = -1

            bits = item.address.split('|')
            mode = 0
            address = 0
            
            if len(bits) == 1:
                address = int(bits[0])
            else:
                address = int(bits[0])
                mode = int(bits[1])
                
            q = -1            
            for fetch in self.fetches:
                q=q+1
                if address >= fetch.address - overhead:
                    if address <= fetch.address + fetch.length + overhead:
                        found = q
                        break
                    
            if found == -1:
                #print "Add New Map"
                fch = fetchset()
                fch.address = -1
                fch.length = 0
                fch.Setup()
                fch.Extend(address,mode,item.code)
                #pnt = fetchsetpoint()
                #pnt.point = item.code
                #pnt.offset = 0
                #pnt.format = mode
                #fch.points.append(pnt)

                self.fetches.append(fch)
            else:           
                self.fetches[found].Extend(address,mode,item.code)

        #for f in self.fetches:
        #    print "Fetching @ " + str(f.address) + " for " + str(f.length) + " on " + str(len(f.points)) + " points, register " + str(f.register);
       
        
class driverfactory:
    def createinstance(self):
        return modbustcpdriver()
            
if __name__ == "__main__":
    sdf = driverfactory()
    base = driverbase.ardidriver()
    base.start(sdf)